import pytest
from tests.mocks import MockOktaClient
from tests.mocks import mock_pause_function
from http import HTTPStatus
import okta.models as models


class TestGroupsResource:
    """
    Integration Tests for the Groups Resource
    """

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        try:
            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Get group using ID
            found_group, _, err = await client.get_group(group.id)
            assert err is None
            assert found_group.id == group.id

            # Delete created group
            _, err = await client.delete_group(group.id)
            assert err is None

            # Ensure group cannot be found again
            found_group, resp, err = await client.get_group(group.id)
            assert err is not None
            assert resp.get_status() == HTTPStatus.NOT_FOUND
            assert found_group is None
        finally:
            # Delete created group if it wasn't deleted during test
            try:
                _, err = await client.delete_group(group.id)
            except Exception:
                pass

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_groups(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        try:
            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            groups_list, resp, err = await client.list_groups()
            assert err is None
            assert not resp.has_next()
            assert next((grp for grp in groups_list if grp.id == group.id))

        finally:
            # Delete created group
            _, err = await client.delete_group(group.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_search_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        try:
            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            query_params_query = {"q": GROUP_NAME}
            groups_list, _, err = await client.list_groups(query_params_query)
            assert err is None
            assert groups_list
            assert len(groups_list) == 1
            assert next((grp for grp in groups_list if grp.id == group.id))

        finally:
            # Delete created group
            _, err = await client.delete_group(group.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        try:
            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Create Updated Group Object
            # Create Group Object
            NEW_GROUP_NAME = "Group-Target-Test NEW"
            new_group_profile = models.GroupProfile({
                "name": NEW_GROUP_NAME
            })
            new_group_obj = models.Group({
                "profile": new_group_profile
            })

            _, _, err = await client.update_group(group.id, new_group_obj)
            assert err is None

            # Verify update worked
            found_group, _, err = await client.get_group(group.id)
            assert err is None
            assert found_group.id == group.id
            assert found_group.profile.name == NEW_GROUP_NAME

        finally:
            # Delete created group
            _, err = await client.delete_group(group.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_remove_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Remove-Group"
        user_profile.email = "John.Doe-Remove-Group@example.com"
        user_profile.login = "John.Doe-Remove-Group@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create query parameters and Create User
            query_params_create = {"activate": "False"}
            user, _, err = await client.create_user(
                create_user_req, query_params_create)
            assert err is None
            assert isinstance(user, models.User)

            # Create Group Object
            GROUP_NAME = "Group-Target-Test"
            group_profile = models.GroupProfile({
                "name": GROUP_NAME
            })
            group_obj = models.Group({
                "profile": group_profile
            })

            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Add user to group
            _, err = await client.add_user_to_group(group.id, user.id)
            assert err is None

            users_in_group, _, err = await client.list_group_users(group.id)
            assert err is None
            assert next((usr for usr in users_in_group if usr.id == user.id))

            # Delete created group
            _, err = await client.delete_group(group.id)
            assert err is None

            # Retrieve group list again to ensure deleted
            groups_list, _, err = await client.list_groups()
            assert err is None
            assert next((grp for grp in groups_list if grp.id ==
                         group.id), None) is None

        finally:
            errors = []
            try:
                _, err = await client.delete_group(group.id)
            except Exception:
                pass

            # Deactivate, then delete created user
            try:
                _, err = await client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_roles_operations(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        try:
            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Create roles
            assign_role_req_ua = models.AssignRoleRequest({
                "type": models.RoleType.USER_ADMIN
            })
            assign_role_req_aa = models.AssignRoleRequest({
                "type": models.RoleType.APP_ADMIN
            })

            ua_role, _, err = await client.assign_role_to_group(
                group.id, assign_role_req_ua)
            assert err is None
            aa_role, _, err = await client.assign_role_to_group(
                group.id, assign_role_req_aa)
            assert err is None

            group_roles, _, err = await client.list_group_assigned_roles(group.id)
            assert err is None
            assert len(group_roles) == 2
            assert next((rle for rle in group_roles if rle.id == ua_role.id))
            assert next((rle for rle in group_roles if rle.id == aa_role.id))

            _, err = await client.remove_role_from_group(group.id, ua_role.id)
            assert err is None

            group_roles, _, err = await client.list_group_assigned_roles(group.id)
            assert err is None
            assert len(group_roles) == 1
            assert next((rle for rle in group_roles if rle.id ==
                         ua_role.id), None) is None
            assert next((rle for rle in group_roles if rle.id == aa_role.id))

        finally:
            # Delete created group
            _, err = await client.delete_group(group.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_users_operations(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Activate"
        user_profile.email = "John.Doe-Activate@example.com"
        user_profile.login = "John.Doe-Activate@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create query parameters and Create User
            query_params_create = {"activate": "False"}
            user, _, err = await client.create_user(
                create_user_req, query_params_create)
            assert err is None
            assert isinstance(user, models.User)

            # Create Group Object
            GROUP_NAME = "Group-Target-Test"
            group_profile = models.GroupProfile({
                "name": GROUP_NAME
            })
            group_obj = models.Group({
                "profile": group_profile
            })

            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Add user to group
            _, err = await client.add_user_to_group(group.id, user.id)
            assert err is None

            users_in_group, _, err = await client.list_group_users(group.id)
            assert err is None
            assert next((usr for usr in users_in_group if usr.id == user.id))

            # Remove user from group
            _, err = await client.remove_user_from_group(group.id, user.id)
            assert err is None

            users_in_group, _, err = await client.list_group_users(group.id)
            assert err is None
            assert len(users_in_group) == 0
            assert next(
                (usr for usr in users_in_group if usr.id == user.id), None) is None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            # Delete created group
            try:
                _, err = await client.delete_group(group.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_rule_operations(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Group-Rule-Ops"
        user_profile.email = "John.Doe-Group-Rule-Ops@example.com"
        user_profile.login = "John.Doe-Group-Rule-Ops@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create query parameters and Create User
            query_params_create = {"activate": "False"}
            user, _, err = await client.create_user(
                create_user_req, query_params_create)
            assert err is None
            assert isinstance(user, models.User)

            # Create Group Object
            GROUP_NAME = "Group-Target-Test-Group-Rule-Ops"
            group_profile = models.GroupProfile({
                "name": GROUP_NAME
            })
            group_obj = models.Group({
                "profile": group_profile
            })

            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Create Group Rule
            last_name = user.profile.last_name
            GROUP_RULE_NAME = "Test-Group-Rule-Group-Rule-Ops"
            GROUP_RULE_TYPE = "group_rule"
            GROUP_RULE_EXP_TYPE = "urn:okta:expression:1.0"
            GROUP_RULE_EXP_VALUE = f"user.lastName==\"{last_name}\""
            group_rule_exp = models.GroupRuleExpression({
                "type": GROUP_RULE_EXP_TYPE,
                "value": GROUP_RULE_EXP_VALUE
            })

            group_rule_cond = models.GroupRuleConditions({
                "expression": group_rule_exp
            })

            group_rule_group_assignment = models.GroupRuleGroupAssignment({
                "groupIds": [group.id]
            })

            group_rule_action = models.GroupRuleAction({
                "assignUserToGroups": group_rule_group_assignment
            })

            group_rule_object = models.GroupRule({
                "actions": group_rule_action,
                "conditions": group_rule_cond,
                "name": GROUP_RULE_NAME,
                "type": GROUP_RULE_TYPE
            })

            group_rule, _, err = await client.create_group_rule(group_rule_object)
            assert err is None
            assert isinstance(group_rule, models.GroupRule)

            # Activate Group Rule
            _, err = await client.activate_group_rule(group_rule.id)
            assert err is None

            # 15 second sleep for backend to update
            mock_pause_function(15)

            users_in_group, _, err = await client.list_group_users(group.id)
            assert err is None
            assert next((usr for usr in users_in_group if usr.id ==
                         user.id), None) is not None

            # Ensure activated rule is in group rules
            group_rules, _, err = await client.list_group_rules()
            assert err is None
            assert next((rule for rule in group_rules if rule.id ==
                         group_rule.id), None) is not None

            # Deactivate rule (to update)
            _, err = await client.deactivate_group_rule(group_rule.id)
            assert err is None

            # Update rule
            # Create new rule
            NEW_GROUP_RULE_NAME = "Test-Group-Rule Updated"
            NEW_GROUP_RULE_EXP_VALUE = "user.lastName==\"BLAHBLAHBLAH\""
            new_group_rule_exp = models.GroupRuleExpression({
                "type": GROUP_RULE_EXP_TYPE,
                "value": NEW_GROUP_RULE_EXP_VALUE
            })

            new_group_rule_cond = models.GroupRuleConditions({
                "expression": new_group_rule_exp
            })

            new_group_rule_group_assignment = models.GroupRuleGroupAssignment({
                "groupIds": [group.id]
            })

            new_group_rule_action = models.GroupRuleAction({
                "assignUserToGroups": new_group_rule_group_assignment
            })

            new_group_rule_object = models.GroupRule({
                "actions": new_group_rule_action,
                "conditions": new_group_rule_cond,
                "name": NEW_GROUP_RULE_NAME,
                "type": GROUP_RULE_TYPE
            })

            new_group_rule, _, err = await client.update_group_rule(
                group_rule.id,
                new_group_rule_object)
            assert err is None

            # Activate updated rule and verify user isn't in group
            _, err = await client.activate_group_rule(new_group_rule.id)
            assert err is None

            # 15 second sleep for backend to update
            mock_pause_function(15)

            users_in_group, _, err = await client.list_group_users(group.id)
            assert err is None
            assert next(
                (usr for usr in users_in_group if usr.id == user.id), None) is None

        finally:
            errors = []
            # Deactivate rule
            try:
                _, err = await client.deactivate_group_rule(new_group_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            # Deactivate, then delete created user
            try:
                _, err = await client.deactivate_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            # Delete created group
            try:
                _, err = await client.delete_group(group.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            # Delete group rules
            try:
                _, err = await client.delete_group_rule(group_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_target_add(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Objects
        GROUP_1_NAME = "Group-Target-Test 1"
        group_1_profile = models.GroupProfile({
            "name": GROUP_1_NAME
        })
        group_1_obj = models.Group({
            "profile": group_1_profile
        })

        GROUP_2_NAME = "Group-Target-Test 2"
        group_2_profile = models.GroupProfile({
            "name": GROUP_2_NAME
        })
        group_2_obj = models.Group({
            "profile": group_2_profile
        })

        try:
            # Create Groups
            group_1, _, err = await client.create_group(group_1_obj)
            assert err is None
            assert isinstance(group_1, models.Group)

            group_2, _, err = await client.create_group(group_2_obj)
            assert err is None
            assert isinstance(group_2, models.Group)

            # Create role and add group targets
            assign_role_req_ua = models.AssignRoleRequest({
                "type": models.RoleType.USER_ADMIN
            })

            ua_role, _, err = await client.assign_role_to_group(
                group_1.id, assign_role_req_ua)
            assert err is None

            _, err = await\
                client.add_group_target_to_group_administrator_role_for_group(
                    group_1.id, ua_role.id, group_2.id)

            # Make sure targets are listed
            groups_list, _, err = await client.list_group_targets_for_group_role(
                group_1.id, ua_role.id)
            assert err is None
            assert next((grp for grp in groups_list if grp.id == group_2.id))

        finally:
            errors = []
            # Delete created groups
            try:
                _, err = await client.delete_group(group_1.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_group(group_2.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_target_remove(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Objects
        GROUP_1_NAME = "Group-Target-Test 1"
        group_1_profile = models.GroupProfile({
            "name": GROUP_1_NAME
        })
        group_1_obj = models.Group({
            "profile": group_1_profile
        })

        GROUP_2_NAME = "Group-Target-Test 2"
        group_2_profile = models.GroupProfile({
            "name": GROUP_2_NAME
        })
        group_2_obj = models.Group({
            "profile": group_2_profile
        })

        GROUP_3_NAME = "Group-Target-Test 3"
        group_3_profile = models.GroupProfile({
            "name": GROUP_3_NAME
        })
        group_3_obj = models.Group({
            "profile": group_3_profile
        })

        try:
            # Create Groups
            group_1, _, err = await client.create_group(group_1_obj)
            assert err is None
            assert isinstance(group_1, models.Group)

            group_2, _, err = await client.create_group(group_2_obj)
            assert err is None
            assert isinstance(group_2, models.Group)

            group_3, _, err = await client.create_group(group_3_obj)
            assert err is None
            assert isinstance(group_3, models.Group)

            # Create role and add group targets
            assign_role_req_ua = models.AssignRoleRequest({
                "type": models.RoleType.USER_ADMIN
            })

            ua_role, _, err = await client.assign_role_to_group(
                group_1.id, assign_role_req_ua)
            assert err is None

            _, err = await\
                client.add_group_target_to_group_administrator_role_for_group(
                    group_1.id, ua_role.id, group_2.id)
            _, err = await\
                client.add_group_target_to_group_administrator_role_for_group(
                    group_1.id, ua_role.id, group_3.id)

            groups_list, _, err = await client.list_group_targets_for_group_role(
                group_1.id, ua_role.id)
            assert err is None
            assert next((grp for grp in groups_list if grp.id == group_2.id))
            assert next((grp for grp in groups_list if grp.id == group_3.id))

            # Remove from 2 and ensure 2 isn't listed
            _, err = await \
                client.remove_group_target_from_group_admin_role_given_to_group(
                    group_1.id, ua_role.id, group_2.id)

            groups_list, _, err = await client.list_group_targets_for_group_role(
                group_1.id, ua_role.id)
            assert err is None
            assert next((grp for grp in groups_list if grp.id ==
                         group_2.id), None) is None
            assert next((grp for grp in groups_list if grp.id == group_3.id))

        finally:
            errors = []
            # Delete created groups
            try:
                _, err = await client.delete_group(group_1.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_group(group_2.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_group(group_3.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_assigned_applications(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Objects
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        try:
            # Create Group
            group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(group, models.Group)

            # Create Application object and Application in Org
            APP_LABEL = "Test Assigned-Applications"
            BUTTON_FIELD = "btn-login"
            PASSWORD_FIELD = "txt-box-password"
            USERNAME_FIELD = "txt-box-username"
            URL = "https://example.com/login.html"
            LOGIN_URL_REGEX = f"^{URL}$"

            swa_app_settings_app = models.SwaApplicationSettingsApplication({
                "buttonField": BUTTON_FIELD,
                "passwordField": PASSWORD_FIELD,
                "usernameField": USERNAME_FIELD,
                "url": URL,
                "loginUrlRegex": LOGIN_URL_REGEX
            })

            swa_app_settings = models.SwaApplicationSettings({
                "app": swa_app_settings_app
            })

            swa_app_obj = models.SwaApplication({
                "label": APP_LABEL,
                "settings": swa_app_settings,
                "signOnMode": models.ApplicationSignOnMode.BROWSER_PLUGIN
            })

            swa_app, _, err = await client.create_application(swa_app_obj)
            assert err is None
            assert isinstance(swa_app, models.SwaApplication)

            # Assign app and group
            assign_ag_req = models.ApplicationGroupAssignment({
                "priority": 0,
                "applicationId": swa_app.id,
                "groupId": group.id
            })

            assign_app_group, _, err = await \
                client.create_application_group_assignment(
                    swa_app.id, group.id, assign_ag_req)
            assert err is None

            # 3 second sleep for backend to update
            mock_pause_function(3)

            # Check assigned apps and ensure created app is found
            assigned_apps, _, err = await \
                client.list_assigned_applications_for_group(group.id)
            assert err is None
            assert assigned_apps is not None
            assert len(assigned_apps) > 0
            assert next((app for app in assigned_apps if app.id == swa_app.id))

        finally:
            errors = []
            # Cleanup app and group created
            try:
                _, err = await client.deactivate_application(swa_app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_application(swa_app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_group(group.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.http_client import HTTPClient
from okta.models.application\
    import Application
from okta.models.provisioning_connection\
    import ProvisioningConnection
from okta.models.csr\
    import Csr
from okta.models.json_web_key\
    import JsonWebKey
from okta.models.application_feature\
    import ApplicationFeature
from okta.models.o_auth_2_scope_consent_grant\
    import OAuth2ScopeConsentGrant
from okta.models.application_group_assignment\
    import ApplicationGroupAssignment
from okta.models.o_auth_2_token\
    import OAuth2Token
from okta.models.app_user\
    import AppUser
from okta.utils import format_url
from okta.api_client import APIClient
from okta.constants import find_app_model


class ApplicationClient(APIClient):
    """
    A Client object for the Application resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_applications(
            self, query_params={},
            keep_empty_params=False
    ):
        """
        Enumerates apps added to your organization with paginat
        ion. A subset of apps can be returned that match a supp
        orted filter expression or query.
        Args:
            query_params {dict}: Map of query parameters for request
            [query_params.q] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
            [query_params.filter] {str}
            [query_params.expand] {str}
            [query_params.includeNonDeleted] {str}
        Returns:
            list: Collection of Application instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Application)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(
                    find_app_model(item["signOnMode"], item["name"])(
                        self.form_response_body(item)
                        )
                    )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_application(
            self, application, query_params={},
            keep_empty_params=False
    ):
        """
        Adds a new application to your Okta organization.
        Args:
            {application}
            query_params {dict}: Map of query parameters for request
            [query_params.activate] {str}
        Returns:
            Application
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(application, dict):
            body = application
        else:
            body = application.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Application)

        if error:
            return (None, response, error)

        try:
            body = response.get_body()
            result = find_app_model(body["signOnMode"], body["name"])(
                self.form_response_body(body)
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Removes an inactive application.
        Args:
            app_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_application(
            self, appId, query_params={},
            keep_empty_params=False
    ):
        """
        Fetches an application from your Okta organization by `
        id`.
        Args:
            app_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            Application
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Application)

        if error:
            return (None, response, error)

        try:
            body = response.get_body()
            result = find_app_model(body["signOnMode"], body["name"])(
                self.form_response_body(body)
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_application(
            self, appId, application,
            keep_empty_params=False
    ):
        """
        Updates an application in your organization.
        Args:
            app_id {str}
            {application}
        Returns:
            Application
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}
            """)

        if isinstance(application, dict):
            body = application
        else:
            body = application.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Application)

        if error:
            return (None, response, error)

        try:
            body = response.get_body()
            result = find_app_model(body["signOnMode"], body["name"])(
                self.form_response_body(body)
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_default_provisioning_connection_for_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Get default Provisioning Connection for application
        Args:
            app_id {str}
        Returns:
            ProvisioningConnection
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/connections/default
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ProvisioningConnection)

        if error:
            return (None, response, error)

        try:
            result = ProvisioningConnection(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def set_default_provisioning_connection_for_application(
            self, appId, provisioning_connection_request, query_params={},
            keep_empty_params=False
    ):
        """
        Set default Provisioning Connection for application
        Args:
            app_id {str}
            {provisioning_connection_request}
            query_params {dict}: Map of query parameters for request
            [query_params.activate] {str}
        Returns:
            ProvisioningConnection
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/connections/default
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(provisioning_connection_request, dict):
            body = provisioning_connection_request
        else:
            body = provisioning_connection_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ProvisioningConnection)

        if error:
            return (None, response, error)

        try:
            result = ProvisioningConnection(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_default_provisioning_connection_for_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Activates the default Provisioning Connection for an ap
        plication.
        Args:
            app_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/connections/default/lifecycle
                /activate
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def deactivate_default_provisioning_connection_for_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Deactivates the default Provisioning Connection for an
        application.
        Args:
            app_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/connections/default/lifecycle
                /deactivate
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_csrs_for_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Enumerates Certificate Signing Requests for an applicat
        ion
        Args:
            app_id {str}
        Returns:
            list: Collection of Csr instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Csr)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Csr(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def generate_csr_for_application(
            self, appId, csr_metadata,
            keep_empty_params=False
    ):
        """
        Generates a new key pair and returns the Certificate Si
        gning Request for it.
        Args:
            app_id {str}
            {csr_metadata}
        Returns:
            Csr
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs
            """)

        if isinstance(csr_metadata, dict):
            body = csr_metadata
        else:
            body = csr_metadata.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Csr)

        if error:
            return (None, response, error)

        try:
            result = Csr(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_csr_from_application(
            self, appId, csrId,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}
        Args:
            app_id {str}
            csr_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_csr_for_application(
            self, appId, csrId,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}
        Args:
            app_id {str}
            csr_id {str}
        Returns:
            Csr
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Csr)

        if error:
            return (None, response, error)

        try:
            result = Csr(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_cer_cert(
            self, appId, csrId, string,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}/lifecycle
        /publish
        Args:
            app_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        if isinstance(string, dict):
            body = string
        else:
            body = string.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-x509-ca-cert"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_binary_cer_cert(
            self, appId, csrId, string,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}/lifecycle
        /publish
        Args:
            app_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        body = HTTPClient.format_binary_data(string)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-x509-ca-cert"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_der_cert(
            self, appId, csrId, string,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}/lifecycle
        /publish
        Args:
            app_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        if isinstance(string, dict):
            body = string
        else:
            body = string.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/pkix-cert"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_binary_der_cert(
            self, appId, csrId, string,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}/lifecycle
        /publish
        Args:
            app_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        body = HTTPClient.format_binary_data(string)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/pkix-cert"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def publish_binary_pem_cert(
            self, appId, csrId, string,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/apps/{appId}/credentials/csrs/{csrId}/lifecycle
        /publish
        Args:
            app_id {str}
            csr_id {str}
            {string}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/csrs/{csrId}
                /lifecycle/publish
            """)

        body = HTTPClient.format_binary_data(string)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-pem-file"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_application_keys(
            self, appId,
            keep_empty_params=False
    ):
        """
        Enumerates key credentials for an application
        Args:
            app_id {str}
        Returns:
            list: Collection of JsonWebKey instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/keys
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(JsonWebKey(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def generate_application_key(
            self, appId, query_params={},
            keep_empty_params=False
    ):
        """
        Generates a new X.509 certificate for an application ke
        y credential
        Args:
            app_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.validityYears] {str}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/keys/generate
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_application_key(
            self, appId, keyId,
            keep_empty_params=False
    ):
        """
        Gets a specific application key credential by kid
        Args:
            app_id {str}
            key_id {str}
        Returns:
            JsonWebKey
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/keys/{keyId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def clone_application_key(
            self, appId, keyId, query_params={},
            keep_empty_params=False
    ):
        """
        Clones a X.509 certificate for an application key crede
        ntial from a source application to target application.
        Args:
            app_id {str}
            key_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.targetAid] {str}
        Returns:
            JsonWebKey
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/credentials/keys/{keyId}/clone
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, JsonWebKey)

        if error:
            return (None, response, error)

        try:
            result = JsonWebKey(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_features_for_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        List Features for application
        Args:
            app_id {str}
        Returns:
            list: Collection of ApplicationFeature instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/features
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ApplicationFeature)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(ApplicationFeature(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_feature_for_application(
            self, appId, name,
            keep_empty_params=False
    ):
        """
        Fetches a Feature object for an application.
        Args:
            app_id {str}
            name {str}
        Returns:
            ApplicationFeature
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/features/{name}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ApplicationFeature)

        if error:
            return (None, response, error)

        try:
            result = ApplicationFeature(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_feature_for_application(
            self, appId, name, capabilities_object,
            keep_empty_params=False
    ):
        """
        Updates a Feature object for an application.
        Args:
            app_id {str}
            name {str}
            {capabilities_object}
        Returns:
            ApplicationFeature
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/features/{name}
            """)

        if isinstance(capabilities_object, dict):
            body = capabilities_object
        else:
            body = capabilities_object.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ApplicationFeature)

        if error:
            return (None, response, error)

        try:
            result = ApplicationFeature(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_scope_consent_grants(
            self, appId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all scope consent grants for the application
        Args:
            app_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            list: Collection of OAuth2ScopeConsentGrant instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/grants
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2ScopeConsentGrant)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OAuth2ScopeConsentGrant(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def grant_consent_to_scope(
            self, appId, o_auth_2_scope_consent_grant,
            keep_empty_params=False
    ):
        """
        Grants consent for the application to request an OAuth
        2.0 Okta scope
        Args:
            app_id {str}
            {o_auth_2_scope_consent_grant}
        Returns:
            OAuth2ScopeConsentGrant
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/grants
            """)

        if isinstance(o_auth_2_scope_consent_grant, dict):
            body = o_auth_2_scope_consent_grant
        else:
            body = o_auth_2_scope_consent_grant.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2ScopeConsentGrant)

        if error:
            return (None, response, error)

        try:
            result = OAuth2ScopeConsentGrant(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_scope_consent_grant(
            self, appId, grantId,
            keep_empty_params=False
    ):
        """
        Revokes permission for the application to request the g
        iven scope
        Args:
            app_id {str}
            grant_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/grants/{grantId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_scope_consent_grant(
            self, appId, grantId, query_params={},
            keep_empty_params=False
    ):
        """
        Fetches a single scope consent grant for the applicatio
        n
        Args:
            app_id {str}
            grant_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            OAuth2ScopeConsentGrant
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/grants/{grantId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2ScopeConsentGrant)

        if error:
            return (None, response, error)

        try:
            result = OAuth2ScopeConsentGrant(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_application_group_assignments(
            self, appId, query_params={},
            keep_empty_params=False
    ):
        """
        Enumerates group assignments for an application.
        Args:
            app_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.q] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
            [query_params.expand] {str}
        Returns:
            list: Collection of ApplicationGroupAssignment instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/groups
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ApplicationGroupAssignment)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(ApplicationGroupAssignment(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_application_group_assignment(
            self, appId, groupId,
            keep_empty_params=False
    ):
        """
        Removes a group assignment from an application.
        Args:
            app_id {str}
            group_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/groups/{groupId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_application_group_assignment(
            self, appId, groupId, query_params={},
            keep_empty_params=False
    ):
        """
        Fetches an application group assignment
        Args:
            app_id {str}
            group_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            ApplicationGroupAssignment
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/groups/{groupId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ApplicationGroupAssignment)

        if error:
            return (None, response, error)

        try:
            result = ApplicationGroupAssignment(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_application_group_assignment(
            self, appId, groupId, application_group_assignment,
            keep_empty_params=False
    ):
        """
        Assigns a group to an application
        Args:
            app_id {str}
            group_id {str}
            {application_group_assignment}
        Returns:
            ApplicationGroupAssignment
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/groups/{groupId}
            """)

        if isinstance(application_group_assignment, dict):
            body = application_group_assignment
        else:
            body = application_group_assignment.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ApplicationGroupAssignment)

        if error:
            return (None, response, error)

        try:
            result = ApplicationGroupAssignment(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Activates an inactive application.
        Args:
            app_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/lifecycle/activate
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def deactivate_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Deactivates an active application.
        Args:
            app_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/lifecycle/deactivate
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def upload_application_logo(
            self, appId, file,
            keep_empty_params=False
    ):
        """
        Update the logo for an application.
        Args:
            app_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/logo
            """)

        body = {}
        headers = {}
        form = {
            "file": file,
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def revoke_o_auth_2_tokens_for_application(
            self, appId,
            keep_empty_params=False
    ):
        """
        Revokes all tokens for the specified application
        Args:
            app_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/tokens
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_o_auth_2_tokens_for_application(
            self, appId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all tokens for the application
        Args:
            app_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of OAuth2Token instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/tokens
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2Token)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OAuth2Token(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_o_auth_2_token_for_application(
            self, appId, tokenId,
            keep_empty_params=False
    ):
        """
        Revokes the specified token for the specified applicati
        on
        Args:
            app_id {str}
            token_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/tokens/{tokenId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_o_auth_2_token_for_application(
            self, appId, tokenId, query_params={},
            keep_empty_params=False
    ):
        """
        Gets a token for the specified application
        Args:
            app_id {str}
            token_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            OAuth2Token
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/tokens/{tokenId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2Token)

        if error:
            return (None, response, error)

        try:
            result = OAuth2Token(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_application_users(
            self, appId, query_params={},
            keep_empty_params=False
    ):
        """
        Enumerates all assigned [application users](#applicatio
        n-user-model) for an application.
        Args:
            app_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.q] {str}
            [query_params.query_scope] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
            [query_params.filter] {str}
            [query_params.expand] {str}
        Returns:
            list: Collection of AppUser instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/users
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, AppUser)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(AppUser(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def assign_user_to_application(
            self, appId, app_user,
            keep_empty_params=False
    ):
        """
        Assigns an user to an application with [credentials](#a
        pplication-user-credentials-object) and an app-specific
        [profile](#application-user-profile-object). Profile m
        appings defined for the application are first applied b
        efore applying any profile properties specified in the
        request.
        Args:
            app_id {str}
            {app_user}
        Returns:
            AppUser
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/users
            """)

        if isinstance(app_user, dict):
            body = app_user
        else:
            body = app_user.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, AppUser)

        if error:
            return (None, response, error)

        try:
            result = AppUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_application_user(
            self, appId, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Removes an assignment for a user from an application.
        Args:
            app_id {str}
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/users/{userId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_application_user(
            self, appId, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Fetches a specific user assignment for application by `
        id`.
        Args:
            app_id {str}
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            AppUser
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/users/{userId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, AppUser)

        if error:
            return (None, response, error)

        try:
            result = AppUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_application_user(
            self, appId, userId, app_user,
            keep_empty_params=False
    ):
        """
        Updates a user's profile for an application
        Args:
            app_id {str}
            user_id {str}
            {app_user}
        Returns:
            AppUser
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/apps/{appId}/users/{userId}
            """)

        if isinstance(app_user, dict):
            body = app_user
        else:
            body = app_user.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, AppUser)

        if error:
            return (None, response, error)

        try:
            result = AppUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.models.profile_mapping\
    import ProfileMapping
from okta.utils import format_url
from okta.api_client import APIClient


class ProfileMappingClient(APIClient):
    """
    A Client object for the ProfileMapping resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_profile_mappings(
            self, query_params={},
            keep_empty_params=False
    ):
        """
        Enumerates Profile Mappings in your organization with p
        agination.
        Args:
            query_params {dict}: Map of query parameters for request
            [query_params.after] {str}
            [query_params.limit] {str}
            [query_params.sourceId] {str}
            [query_params.targetId] {str}
        Returns:
            list: Collection of ProfileMapping instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/mappings
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ProfileMapping)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(ProfileMapping(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_profile_mapping(
            self, mappingId,
            keep_empty_params=False
    ):
        """
        Fetches a single Profile Mapping referenced by its ID.
        Args:
            mapping_id {str}
        Returns:
            ProfileMapping
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/mappings/{mappingId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ProfileMapping)

        if error:
            return (None, response, error)

        try:
            result = ProfileMapping(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_profile_mapping(
            self, mappingId, profile_mapping,
            keep_empty_params=False
    ):
        """
        Updates an existing Profile Mapping by adding, updating
        , or removing one or many Property Mappings.
        Args:
            mapping_id {str}
            {profile_mapping}
        Returns:
            ProfileMapping
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/mappings/{mappingId}
            """)

        if isinstance(profile_mapping, dict):
            body = profile_mapping
        else:
            body = profile_mapping.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ProfileMapping)

        if error:
            return (None, response, error)

        try:
            result = ProfileMapping(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestPoliciesResource:
    """
    Integration Tests for the Policies Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sign_on_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type

        finally:
            # Delete
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_password_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.PasswordPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.PasswordPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.PASSWORD

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type

        finally:
            # Delete
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.Policy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "type": models.PolicyType.OKTA_SIGN_ON,
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type

        finally:
            # Delete
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sign_on_policy_with_group_conditions(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group
        GROUP_NAME = "Group-Target-Test"
        group_obj = models.Group({
            "profile": models.GroupProfile({
                "name": GROUP_NAME
            })
        })

        try:
            created_group, _, err = await client.create_group(group_obj)
            assert err is None
            assert isinstance(created_group, models.Group)

            # Create Policy & Conditions
            policy_model = models.OktaSignOnPolicy({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
                "status": "ACTIVE",
                "description": "Test policy applies for tests only"
            })

            policy_conditions = models.OktaSignOnPolicyConditions({
                "people": models.PolicyPeopleCondition({
                    "groups": models.GroupCondition({
                        "include": [created_group.id]
                    })
                })
            })

            policy_model.conditions = policy_conditions

            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON
            assert len(created_policy.conditions.people.groups.include) == 1
            assert created_group.id in \
                created_policy.conditions.people.groups.include

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type
            assert len(retrieved_policy.conditions.people.groups.include) == 1
            assert created_group.id in \
                retrieved_policy.conditions.people.groups.include

        finally:
            errors = []
            # Delete Policy + Group
            try:
                _, err = await client.delete_policy(created_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_group(created_group.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_policies_by_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policies
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test sign on policy applies for tests only"
        })

        policy_model_2 = models.PasswordPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password",
            "status": "ACTIVE",
            "description": "Test password policy applies for tests only"
        })

        try:
            created_oss_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_oss_policy, models.OktaSignOnPolicy)
            assert created_oss_policy.name == policy_model.name
            assert created_oss_policy.description == policy_model.description
            assert created_oss_policy.status == "ACTIVE"
            assert created_oss_policy.type == models.PolicyType.OKTA_SIGN_ON

            created_pw_policy, _, err = await client.create_policy(policy_model_2)
            assert err is None
            assert isinstance(created_pw_policy, models.PasswordPolicy)
            assert created_pw_policy.name == policy_model_2.name
            assert created_pw_policy.description == policy_model_2.description
            assert created_pw_policy.status == "ACTIVE"
            assert created_pw_policy.type == models.PolicyType.PASSWORD

            # List by type
            query_params_list_sign_on = {"type": "OKTA_SIGN_ON"}
            sign_on_policies, _, err = await \
                client.list_policies(query_params_list_sign_on)
            assert err is None
            assert isinstance(sign_on_policies, list)
            assert next((plcy for plcy in sign_on_policies
                         if plcy.id == created_oss_policy.id))

            query_params_list_pw = {"type": "PASSWORD"}
            sign_on_policies, _, err = await \
                client.list_policies(query_params_list_pw)
            assert err is None
            assert isinstance(sign_on_policies, list)
            assert next((plcy for plcy in sign_on_policies
                         if plcy.id == created_pw_policy.id))

        finally:
            errors = []
            # Delete
            try:
                _, err = await client.delete_policy(created_oss_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_policy(created_pw_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type

            # Delete
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

            # Retrieve
            retrieved_policy, resp, err = await \
                client.get_policy(created_policy.id)
            assert err is not None
            assert isinstance(err, OktaAPIError)
            assert resp.get_status() == HTTPStatus.NOT_FOUND
            assert retrieved_policy is None
        finally:
            try:
                _, err = await client.delete_policy(created_policy.id)
            except Exception:
                pass

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type

            # Update
            NEW_NAME = policy_model.name + "UPDATED"
            NEW_DESC = policy_model.description + "UPDATED"
            created_policy.name = NEW_NAME
            created_policy.description = NEW_DESC
            updated_policy, _, err = await client.update_policy(
                created_policy.id, created_policy
            )
            assert err is None
            assert updated_policy.id == created_policy.id
            assert updated_policy.name == NEW_NAME
            assert updated_policy.description == NEW_DESC

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == NEW_NAME
            assert retrieved_policy.description == NEW_DESC
            assert retrieved_policy.status == created_policy.status
            assert retrieved_policy.type == created_policy.type

        finally:
            # Delete
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Deactivate
            _, err = await client.deactivate_policy(created_policy.id)
            assert err is None

            # Retrieve
            retrieved_policy, _, err = await client.get_policy(created_policy.id)
            assert err is None
            assert retrieved_policy.id == created_policy.id
            assert retrieved_policy.name == created_policy.name
            assert retrieved_policy.description == created_policy.description
            assert retrieved_policy.status == "INACTIVE"
            assert retrieved_policy.type == created_policy.type

            # Reactivate
            _, err = await client.activate_policy(created_policy.id)
            assert err is None

        finally:
            # Delete
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

    """
    POLICY RULES TESTS BELOW
    """
    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sign_on_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Create Policy Rule
            policy_rule_model = models.OktaSignOnPolicyRule({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
                "actions": models.OktaSignOnPolicyRuleActions({
                    "signon": models.OktaSignOnPolicyRuleSignonActions({
                        "access": "ALLOW",
                        "requireFactor": "false",
                        "rememberDeviceByDefault": "false",
                        "session":
                        models.OktaSignOnPolicyRuleSignonSessionActions({
                            "usePersistentCookie": "false",
                            "maxSessionIdleMinutes": 720,
                            "maxSessionLifetimeMinutes": 0
                        })
                    })
                }),
                "conditions": models.OktaSignOnPolicyRuleConditions({
                    "authContext": models.PolicyRuleAuthContextCondition({
                        "authType": "ANY"
                    })
                })
            })
            created_policy_rule, _, err = await client.create_policy_rule(
                created_policy.id, policy_rule_model
            )
            assert err is None
            assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
            assert created_policy_rule.name == policy_rule_model.name
            assert created_policy_rule.type == "SIGN_ON"
            assert created_policy_rule.actions.signon.access == "ALLOW"
            assert created_policy_rule.actions.signon.require_factor is False
            assert created_policy_rule.actions.signon.remember_device_by_default\
                is False
            assert created_policy_rule.actions.signon.session.\
                use_persistent_cookie is False
            assert created_policy_rule.actions.signon.session.\
                max_session_idle_minutes == 720
            assert created_policy_rule.actions.signon.session.\
                max_session_lifetime_minutes == 0
            assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        finally:
            errors = []
            # Delete
            try:
                _, err = await \
                    client.delete_policy_rule(
                        created_policy.id, created_policy_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_policy(created_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_password_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.PasswordPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.PasswordPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.PASSWORD

            # Create Policy Rule
            policy_rule_model = models.PasswordPolicyRule({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password-Rule",
                "actions": models.PasswordPolicyRuleActions({
                    "passwordChange": models.PasswordPolicyRuleAction({
                        "access": "ALLOW"
                    }),
                    "selfServicePasswordReset": models.PasswordPolicyRuleAction({
                        "access": "ALLOW"
                    }),
                    "selfServiceUnlock": models.PasswordPolicyRuleAction({
                        "access": "ALLOW"
                    })
                }),
                "conditions": models.PasswordPolicyRuleConditions({
                    "people": models.PolicyPeopleCondition({
                        "users": models.UserCondition({
                            "exclude": []
                        })
                    }),
                    "network": models.PolicyNetworkCondition({
                        "connection": "ANYWHERE"
                    })
                })
            })
            created_policy_rule, _, err = await client.create_policy_rule(
                created_policy.id, policy_rule_model
            )
            assert err is None
            assert isinstance(created_policy_rule, models.PasswordPolicyRule)
            assert created_policy_rule.name == policy_rule_model.name
            assert created_policy_rule.type == "PASSWORD"
            assert created_policy_rule.status == "ACTIVE"
            assert created_policy_rule.conditions.people.users.exclude == []
            assert created_policy_rule.conditions.network.connection ==\
                "ANYWHERE"
            assert created_policy_rule.actions.password_change.access == "ALLOW"
            assert created_policy_rule.actions.self_service_password_reset\
                .access == "ALLOW"
            assert created_policy_rule.actions.self_service_unlock\
                .access == "ALLOW"

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert isinstance(retrieved_policy_rule, models.PasswordPolicyRule)

        finally:
            errors = []
            # Delete
            try:
                _, err = await \
                    client.delete_policy_rule(
                        created_policy.id, created_policy_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_policy(created_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Create Policy Rule
            policy_rule_model = models.OktaSignOnPolicyRule({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
                "actions": models.OktaSignOnPolicyRuleActions({
                    "signon": models.OktaSignOnPolicyRuleSignonActions({
                        "access": "ALLOW",
                        "requireFactor": "false",
                        "rememberDeviceByDefault": "false",
                        "session":
                        models.OktaSignOnPolicyRuleSignonSessionActions({
                            "usePersistentCookie": "false",
                            "maxSessionIdleMinutes": 720,
                            "maxSessionLifetimeMinutes": 0
                        })
                    })
                }),
                "conditions": models.OktaSignOnPolicyRuleConditions({
                    "authContext": models.PolicyRuleAuthContextCondition({
                        "authType": "ANY"
                    })
                })
            })
            created_policy_rule, _, err = await client.create_policy_rule(
                created_policy.id, policy_rule_model
            )
            assert err is None
            assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
            assert created_policy_rule.name == policy_rule_model.name
            assert created_policy_rule.type == "SIGN_ON"
            assert created_policy_rule.actions.signon.access == "ALLOW"
            assert created_policy_rule.actions.signon.require_factor is False
            assert created_policy_rule.actions.signon.remember_device_by_default\
                is False
            assert created_policy_rule.actions.signon.session.\
                use_persistent_cookie is False
            assert created_policy_rule.actions.signon.session.\
                max_session_idle_minutes == 720
            assert created_policy_rule.actions.signon.session.\
                max_session_lifetime_minutes == 0
            assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

            # Update
            created_policy_rule.name = policy_rule_model.name + "UPDATED"
            created_policy_rule.conditions.network.connection = "ANYWHERE"
            updated_policy_rule, _, err = await client.update_policy_rule(
                created_policy.id, created_policy_rule.id, created_policy_rule
            )
            assert err is None
            assert isinstance(updated_policy_rule, models.OktaSignOnPolicyRule)
            assert updated_policy_rule.name == created_policy_rule.name
            assert updated_policy_rule.conditions.network.connection ==\
                created_policy_rule.conditions.network.connection

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert retrieved_policy_rule.conditions.network.connection ==\
                created_policy_rule.conditions.network.connection
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        finally:
            errors = []
            # Delete
            try:
                _, err = await \
                    client.delete_policy_rule(
                        created_policy.id, created_policy_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_policy(created_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_policy_rules(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Create Policy Rule
            policy_rule_model = models.OktaSignOnPolicyRule({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
                "actions": models.OktaSignOnPolicyRuleActions({
                    "signon": models.OktaSignOnPolicyRuleSignonActions({
                        "access": "ALLOW",
                        "requireFactor": "false",
                        "rememberDeviceByDefault": "false",
                        "session":
                        models.OktaSignOnPolicyRuleSignonSessionActions({
                            "usePersistentCookie": "false",
                            "maxSessionIdleMinutes": 720,
                            "maxSessionLifetimeMinutes": 0
                        })
                    })
                }),
                "conditions": models.OktaSignOnPolicyRuleConditions({
                    "authContext": models.PolicyRuleAuthContextCondition({
                        "authType": "ANY"
                    })
                })
            })
            created_policy_rule, _, err = await client.create_policy_rule(
                created_policy.id, policy_rule_model
            )
            assert err is None
            assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
            assert created_policy_rule.name == policy_rule_model.name
            assert created_policy_rule.type == "SIGN_ON"
            assert created_policy_rule.actions.signon.access == "ALLOW"
            assert created_policy_rule.actions.signon.require_factor is False
            assert created_policy_rule.actions.signon.remember_device_by_default\
                is False
            assert created_policy_rule.actions.signon.session.\
                use_persistent_cookie is False
            assert created_policy_rule.actions.signon.session.\
                max_session_idle_minutes == 720
            assert created_policy_rule.actions.signon.session.\
                max_session_lifetime_minutes == 0
            assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

            # List
            policy_rules, _, err = await \
                client.list_policy_rules(created_policy.id)
            assert err is None
            assert isinstance(policy_rules, list)
            assert next((plcy_rule for plcy_rule in policy_rules
                         if plcy_rule.id == created_policy_rule.id))
            assert len(policy_rules) == 1

        finally:
            errors = []
            # Delete
            try:
                _, err = await \
                    client.delete_policy_rule(
                        created_policy.id, created_policy_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_policy(created_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_policy_rules(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Create Policy Rule
            policy_rule_model = models.OktaSignOnPolicyRule({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
                "actions": models.OktaSignOnPolicyRuleActions({
                    "signon": models.OktaSignOnPolicyRuleSignonActions({
                        "access": "ALLOW",
                        "requireFactor": "false",
                        "rememberDeviceByDefault": "false",
                        "session":
                        models.OktaSignOnPolicyRuleSignonSessionActions({
                            "usePersistentCookie": "false",
                            "maxSessionIdleMinutes": 720,
                            "maxSessionLifetimeMinutes": 0
                        })
                    })
                }),
                "conditions": models.OktaSignOnPolicyRuleConditions({
                    "authContext": models.PolicyRuleAuthContextCondition({
                        "authType": "ANY"
                    })
                })
            })
            created_policy_rule, _, err = await client.create_policy_rule(
                created_policy.id, policy_rule_model
            )
            assert err is None
            assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
            assert created_policy_rule.name == policy_rule_model.name
            assert created_policy_rule.type == "SIGN_ON"
            assert created_policy_rule.actions.signon.access == "ALLOW"
            assert created_policy_rule.actions.signon.require_factor is False
            assert created_policy_rule.actions.signon.remember_device_by_default\
                is False
            assert created_policy_rule.actions.signon.session.\
                use_persistent_cookie is False
            assert created_policy_rule.actions.signon.session.\
                max_session_idle_minutes == 720
            assert created_policy_rule.actions.signon.session.\
                max_session_lifetime_minutes == 0
            assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

            # Delete
            _, err = await \
                client.delete_policy_rule(
                    created_policy.id, created_policy_rule.id)
            assert err is None
            _, err = await client.delete_policy(created_policy.id)
            assert err is None

            # Retrieve
            retrieved_policy_rule, resp, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is not None
            assert isinstance(err, OktaAPIError)
            assert resp.get_status() == HTTPStatus.NOT_FOUND
            assert retrieved_policy_rule is None
        finally:
            # Delete
            try:
                _, err = await \
                    client.delete_policy_rule(
                        created_policy.id, created_policy_rule.id)
            except Exception:
                pass
            try:
                _, err = await client.delete_policy(created_policy.id)
            except Exception:
                pass

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        try:
            created_policy, _, err = await client.create_policy(policy_model)
            assert err is None
            assert isinstance(created_policy, models.OktaSignOnPolicy)
            assert created_policy.name == policy_model.name
            assert created_policy.description == policy_model.description
            assert created_policy.status == "ACTIVE"
            assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

            # Create Policy Rule
            policy_rule_model = models.OktaSignOnPolicyRule({
                "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
                "actions": models.OktaSignOnPolicyRuleActions({
                    "signon": models.OktaSignOnPolicyRuleSignonActions({
                        "access": "ALLOW",
                        "requireFactor": "false",
                        "rememberDeviceByDefault": "false",
                        "session":
                        models.OktaSignOnPolicyRuleSignonSessionActions({
                            "usePersistentCookie": "false",
                            "maxSessionIdleMinutes": 720,
                            "maxSessionLifetimeMinutes": 0
                        })
                    })
                }),
                "conditions": models.OktaSignOnPolicyRuleConditions({
                    "authContext": models.PolicyRuleAuthContextCondition({
                        "authType": "ANY"
                    })
                })
            })
            created_policy_rule, _, err = await client.create_policy_rule(
                created_policy.id, policy_rule_model
            )
            assert err is None
            assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
            assert created_policy_rule.name == policy_rule_model.name
            assert created_policy_rule.type == "SIGN_ON"
            assert created_policy_rule.status == "ACTIVE"
            assert created_policy_rule.actions.signon.access == "ALLOW"
            assert created_policy_rule.actions.signon.require_factor is False
            assert created_policy_rule.actions.signon.remember_device_by_default\
                is False
            assert created_policy_rule.actions.signon.session.\
                use_persistent_cookie is False
            assert created_policy_rule.actions.signon.session.\
                max_session_idle_minutes == 720
            assert created_policy_rule.actions.signon.session.\
                max_session_lifetime_minutes == 0
            assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert retrieved_policy_rule.status == "ACTIVE"
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

            # Deactivate
            _, err = await client.deactivate_policy_rule(
                created_policy.id, created_policy_rule.id
            )
            assert err is None

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert retrieved_policy_rule.status == "INACTIVE"
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

            # Reactivate
            _, err = await client.activate_policy_rule(
                created_policy.id, created_policy_rule.id
            )
            assert err is None

            # Retrieve
            retrieved_policy_rule, _, err = await client.get_policy_rule(
                created_policy.id, created_policy_rule.id)
            assert err is None
            assert retrieved_policy_rule.id == created_policy_rule.id
            assert retrieved_policy_rule.name == created_policy_rule.name
            assert retrieved_policy_rule.status == "ACTIVE"
            assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        finally:
            errors = []
            # Delete
            try:
                _, err = await \
                    client.delete_policy_rule(
                        created_policy.id, created_policy_rule.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_policy(created_policy.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.models.user_type\
    import UserType
from okta.utils import format_url
from okta.api_client import APIClient


class UserTypeClient(APIClient):
    """
    A Client object for the UserType resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_user_types(
            self,
            keep_empty_params=False
    ):
        """
        Fetches all User Types in your org
        Args:
        Returns:
            list: Collection of UserType instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/meta/types/user
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserType)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(UserType(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_user_type(
            self, user_type,
            keep_empty_params=False
    ):
        """
        Creates a new User Type. A default User Type is automat
        ically created along with your org, and you may add ano
        ther 9 User Types for a maximum of 10.
        Args:
            {user_type}
        Returns:
            UserType
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/meta/types/user
            """)

        if isinstance(user_type, dict):
            body = user_type
        else:
            body = user_type.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserType)

        if error:
            return (None, response, error)

        try:
            result = UserType(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_user_type(
            self, typeId,
            keep_empty_params=False
    ):
        """
        Deletes a User Type permanently. This operation is not
        permitted for the default type, nor for any User Type t
        hat has existing users
        Args:
            type_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/meta/types/user/{typeId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_user_type(
            self, typeId,
            keep_empty_params=False
    ):
        """
        Fetches a User Type by ID. The special identifier `defa
        ult` may be used to fetch the default User Type.
        Args:
            type_id {str}
        Returns:
            UserType
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/meta/types/user/{typeId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserType)

        if error:
            return (None, response, error)

        try:
            result = UserType(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_user_type(
            self, typeId, user_type,
            keep_empty_params=False
    ):
        """
        Updates an existing User Type
        Args:
            type_id {str}
            {user_type}
        Returns:
            UserType
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/meta/types/user/{typeId}
            """)

        if isinstance(user_type, dict):
            body = user_type
        else:
            body = user_type.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserType)

        if error:
            return (None, response, error)

        try:
            result = UserType(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def replace_user_type(
            self, typeId, user_type,
            keep_empty_params=False
    ):
        """
        Replace an existing User Type
        Args:
            type_id {str}
            {user_type}
        Returns:
            UserType
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/meta/types/user/{typeId}
            """)

        if isinstance(user_type, dict):
            body = user_type
        else:
            body = user_type.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserType)

        if error:
            return (None, response, error)

        try:
            result = UserType(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



import pytest
from tests.mocks import MockOktaClient
from okta.models import Domain, DomainListResponse


class TestDomainResource:
    """
    Integration Tests for the Domain Resource
    """

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_domains(self, fs):
        client = MockOktaClient(fs)
        domain_list_resp, _, err = await client.list_domains()
        assert isinstance(domain_list_resp, DomainListResponse)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_and_delete_domain(self, fs):
        client = MockOktaClient(fs)
        domain_config = {
            "domain": "login.example.com",
            "certificateSourceType": "MANUAL"
        }
        domain, _, err = await client.create_domain(domain_config)
        assert err is None
        try:
            assert isinstance(domain, Domain)
            assert domain.id
        finally:
            _, err = await client.delete_domain(domain.id)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_domain(self, fs):
        client = MockOktaClient(fs)
        domain_config = {
            "domain": "login.example.com",
            "certificateSourceType": "MANUAL"
        }
        domain, _, err = await client.create_domain(domain_config)
        assert err is None
        try:
            get_domain, _, err = await client.get_domain(domain.id)
            assert isinstance(domain, Domain)
            assert domain.id == get_domain.id
            assert domain.domain == get_domain.domain
        finally:
            _, err = await client.delete_domain(domain.id)
            assert err is None



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.models.threat_insight_configuration\
    import ThreatInsightConfiguration
from okta.utils import format_url
from okta.api_client import APIClient


class ThreatInsightClient(APIClient):
    """
    A Client object for the ThreatInsight resource.
    """

    def __init__(self):
        self._base_url = ""

    async def get_current_configuration(
            self,
            keep_empty_params=False
    ):
        """
        Gets current ThreatInsight configuration
        Args:
        Returns:
            ThreatInsightConfiguration
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/threats/configuration
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ThreatInsightConfiguration)

        if error:
            return (None, response, error)

        try:
            result = ThreatInsightConfiguration(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_configuration(
            self, threat_insight_configuration,
            keep_empty_params=False
    ):
        """
        Updates ThreatInsight configuration
        Args:
            {threat_insight_configuration}
        Returns:
            ThreatInsightConfiguration
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/threats/configuration
            """)

        if isinstance(threat_insight_configuration, dict):
            body = threat_insight_configuration
        else:
            body = threat_insight_configuration.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ThreatInsightConfiguration)

        if error:
            return (None, response, error)

        try:
            result = ThreatInsightConfiguration(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



import pickle
import pytest
import okta.models as models


def test_attribute_case():
    """Verify UserProfile can handle different case for basic and custom attributes."""
    first_name = 'TestFirstName'
    last_name = 'TestLastName'
    custom_attr = 'TestCustomAttr'
    config = {'firstName': first_name,
              'lastName': last_name,
              'CustomAttr': custom_attr}
    profile = models.UserProfile(config)
    assert profile.first_name == first_name
    assert profile.firstName == first_name
    assert profile.last_name == last_name
    assert profile.lastName == last_name
    assert profile.display_name is None
    assert profile.displayName is None
    assert profile.CustomAttr == custom_attr

    display_name = 'TestName'
    profile.displayName = display_name
    assert profile.display_name == display_name

    new_display_name = 'NewTestName'
    profile.display_name = new_display_name
    assert profile.displayName == new_display_name

    new_custom_attr = 'NewCustomAttr'
    profile.CustomAttr = new_custom_attr

    req_format = profile.request_format()
    assert req_format['CustomAttr'] == new_custom_attr
    assert req_format['firstName'] == first_name
    assert req_format['lastName'] == last_name
    assert req_format['displayName'] == new_display_name


def test_profile_serializing():
    """Verify UserProfile can be serialized and deserialized with pickle library."""
    first_name = 'TestFirstName'
    last_name = 'TestLastName'
    custom_attr = 'TestCustomAttr'
    config = {'firstName': first_name,
              'lastName': last_name,
              'CustomAttr': custom_attr}
    profile = models.UserProfile(config)
    pickled_profile = pickle.dumps(profile)
    unpickled_profile = pickle.loads(pickled_profile)
    assert isinstance(unpickled_profile, models.UserProfile)
    assert profile.as_dict() == unpickled_profile.as_dict()



import pytest
import okta.models as models
import time
from datetime import datetime
from tests.mocks import MockOktaClient
from okta.errors.okta_api_error import OktaAPIError
from okta.constants import DATETIME_FORMAT


class TestUsersResource:
    """
    Integration Tests for the Users Resource.
    """

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })

        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Get"
        user_profile.email = "John.Doe-Get@example.com"
        user_profile.login = "John.Doe-Get@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        query_params_create = {"activate": "False"}

        try:
            # Create User
            user, resp, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            got_user, resp, err = await test_client.get_user(user.id)
            assert got_user.id == user.id

            got_user_with_string, resp, err = await test_client.get_user(
                user_profile.login)
            assert got_user_with_string.id == user.id
            assert user.profile.first_name == user_profile.first_name
            assert user.profile.last_name == user_profile.last_name
            assert user.profile.email == user_profile.email
            assert user.profile.login == user_profile.login

            user_id = user.id

            # Deactivate, then delete created user
            _, err = await test_client.deactivate_or_delete_user(user.id)
            assert err is None

            _, err = await test_client.deactivate_or_delete_user(user.id)
            assert err is None

            user, _, err = await test_client.get_user(user_id)
            assert user is None
            assert isinstance(err, OktaAPIError)
        finally:
            # Clean up
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
            except Exception:
                pass

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
            except Exception:
                pass

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Activate"
        user_profile.email = "John.Doe-Activate@example.com"
        user_profile.login = "John.Doe-Activate@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create query parameters and Create User
            query_params_create = {"activate": "False"}
            user, resp, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Create query parameters and Activate User
            query_params_activate = {"sendEmail": "False"}
            token, resp, err = await test_client.activate_user(
                user.id, query_params_activate)
            assert err is None
            assert token is not None
            # Ensure correct Object is returned
            assert isinstance(token, models.UserActivationToken)

            # Create query parameters and List Users
            query_params_list = {"filter": "status eq \"ACTIVE\""}
            users, resp, err = await test_client.list_users(query_params_list)
            assert err is None
            # Ensure user is in list
            assert next((usr for usr in users if usr.id ==
                         user.id), None) is not None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_user_profile(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Update-Profile"
        user_profile.email = "John.Doe-Update-Profile@example.com"
        user_profile.login = "John.Doe-Update-Profile@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create query parameters and Create User
            query_params_create = {"activate": "False"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Craft new profile and get user object
            new_profile = user.profile
            NICK_NAME = "JD"
            new_profile.nick_name = NICK_NAME
            updated_user = models.User({"profile": new_profile})

            # Update User with new details
            new_user, _, err = await test_client.update_user(user.id, updated_user)
            assert err is None
            assert new_user is not None

            # Retrieve user and ensure details are updated
            got_user, _, err = await test_client.get_user(user.id)
            assert err is None
            assert got_user.id == user.id
            assert got_user.profile.nick_name == NICK_NAME

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_suspend_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Suspend"
        user_profile.email = "John.Doe-Suspend@example.com"
        user_profile.login = "John.Doe-Suspend@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Suspend User
            _, err = await test_client.suspend_user(user.id)
            assert err is None

            # Create query params and List Suspended Users
            query_params_list = {"filter": "status eq \"SUSPENDED\""}
            users, resp, err = await test_client.list_users(query_params_list)
            assert err is None
            # Ensure created user is in list
            assert next((usr for usr in users if usr.id ==
                         user.id), None) is not None

            # Unsuspend User
            _, err = await test_client.unsuspend_user(user.id)
            assert err is None
            # Create query params and List Active Users
            query_params_list = {"filter": "status eq \"ACTIVE\""}
            users, resp, err = await test_client.list_users(query_params_list)
            assert err is None
            # Ensure created user is in list
            assert next((usr for usr in users if usr.id ==
                         user.id), None) is not None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_change_user_password(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Change-PW"
        user_profile.email = "John.Doe-Change-PW@example.com"
        user_profile.login = "John.Doe-Change-PW@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Give time before changing password
            time.sleep(1)

            # Create new Password
            NEW_PASSWORD = "2ANewPassword!"
            new_password = models.PasswordCredential({
                "value": NEW_PASSWORD
            })

            change_pw_req = models.ChangePasswordRequest({
                "oldPassword": password,
                "newPassword": new_password
            })

            # change password and retrieve user after
            _, _, err = await test_client.change_password(user.id, change_pw_req)
            assert err is None

            got_user, resp, err = await test_client.get_user(user.id)
            assert got_user.id == user.id
            assert err is None
            # Verify password in updated user obj, changed after original
            DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
            assert datetime.strptime(got_user.password_changed, DATETIME_FORMAT)\
                > datetime.strptime(user.password_changed, DATETIME_FORMAT)

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_reset_password_link_for_user(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Reset-PW-URL"
        user_profile.email = "John.Doe-Reset-PW-URL@example.com"
        user_profile.login = "John.Doe-Reset-PW-URL@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)

            # Create Query Parameters and get Reset Password Token
            query_params_reset_pw = {"sendEmail": "False"}
            reset_pw_token, _, err = await test_client.reset_password(
                user.id, query_params_reset_pw)
            assert err is None
            assert isinstance(reset_pw_token, models.ResetPasswordToken)
            assert reset_pw_token is not None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_expire_password_get_temporary(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Expire-PW-URL"
        user_profile.email = "John.Doe-Expire-PW-URL@example.com"
        user_profile.login = "John.Doe-Expire-PW-URL@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Create Query Parameters and get Temporary Password
            temporary_pw, _, err = await\
                test_client.expire_password_and_get_temporary_password(user.id)
            assert err is None
            assert isinstance(temporary_pw, models.TempPassword)
            assert temporary_pw is not None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_change_recovery_question(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Change-Recovery"
        user_profile.email = "John.Doe-Change-Recovery@example.com"
        user_profile.login = "John.Doe-Change-Recovery@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Create Recovery Question
            NEW_RECOVERY_Q = "What's your favourite security question?"
            NEW_RECOVERY_A = "This One!"

            new_recovery_question = models.RecoveryQuestionCredential({
                "question": NEW_RECOVERY_Q,
                "answer": NEW_RECOVERY_A
            })

            new_user_creds = models.UserCredentials({
                "password": password,
                "recoveryQuestion": new_recovery_question
            })

            # Change Recovery Question
            updated_user_creds, _, err = await\
                test_client.change_recovery_question(user.id, new_user_creds)
            assert err is None
            assert isinstance(updated_user_creds, models.UserCredentials)

            # Create New Password
            NEW_PASSWORD = password.value[::-1]  # reverse string
            new_pw = models.PasswordCredential({
                "value": NEW_PASSWORD
            })
            recovery_answer = models.RecoveryQuestionCredential({
                "answer": NEW_RECOVERY_A
            })

            change_pw_user_creds = models.UserCredentials({
                "password": new_pw,
                "recoveryQuestion": recovery_answer
            })

            # Create New Password using Recovery Question answer
            _, _, err = await test_client.forgot_password_set_new_password(
                user.id, change_pw_user_creds)
            assert err is None

            # Get User and verify password was changed
            got_user, _, err = await test_client.get_user(user.id)
            assert got_user.id == user.id
            assert err is None
            DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
            assert datetime.strptime(got_user.password_changed, DATETIME_FORMAT)\
                > datetime.strptime(user.password_changed, DATETIME_FORMAT)

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_assign_user_to_role(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Assign-User-Role"
        user_profile.email = "John.Doe-Assign-User-Role@example.com"
        user_profile.login = "John.Doe-Assign-User-Role@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Create Assign Role Request with Roletype Enum
            USER_ADMIN = models.RoleType.USER_ADMIN
            assign_role_req = models.AssignRoleRequest({
                "type": USER_ADMIN
            })

            # Assign Role to User
            _, _, err = await test_client.assign_role_to_user(user.id,
                                                              assign_role_req)
            assert err is None

            # Get Roles for user and ensure role assigned is found
            roles, _, err = await test_client.list_assigned_roles_for_user(user.id)
            found_role = next((role for role in roles if role.type ==
                               USER_ADMIN), None)
            assert found_role is not None

            # Remove assigned role from user
            _, err = await test_client.remove_role_from_user(user.id,
                                                             found_role.id)
            assert err is None

            # Get Roles for user and ensure role assigned is NOT found
            roles, _, err = await test_client.list_assigned_roles_for_user(
                user.id)
            found_role = next((role for role in roles if role.type ==
                               USER_ADMIN), None)
            assert found_role is None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_user_group_target_to_role(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Group-Target-Role-Assign"
        user_profile.email = "John.Doe-Group-Target-Role-Assign@example.com"
        user_profile.login = "John.Doe-Group-Target-Role-Assign@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        try:
            # Create Query Parameters and Create User
            query_params_create = {"activate": "True"}
            user, _, err = await test_client.create_user(
                create_user_req, query_params_create)
            assert err is None

            # Create Group Object
            NEW_GROUP_NAME = "Group-Target-Test-Assign"
            new_group_profile = models.GroupProfile({
                "name": NEW_GROUP_NAME
            })
            new_group = models.Group({
                "profile": new_group_profile
            })

            # Create Group
            group, _, err = await test_client.create_group(new_group)
            assert err is None

            # Create request to assign role to user
            USER_ADMIN = models.RoleType.USER_ADMIN
            assign_role_req = models.AssignRoleRequest({
                "type": USER_ADMIN
            })

            # Assign Role to User
            user_role, _, err = await test_client.assign_role_to_user(
                user.id, assign_role_req)
            assert err is None

            # Add Group Target to the Role
            _, err = await test_client.add_group_target_to_role(
                user.id, user_role.id, group.id)
            assert err is None

            # Retrieve group targets for role and ensure added one is there
            groups, _, err = await test_client.list_group_targets_for_role(
                user.id, user_role.id)
            assert next((grp for grp in groups if grp.id ==
                         group.id), None) is not None

            # Create another group to add
            NEW_GROUP_NAME = "Temp-Group-Target-Test-Assign"
            new_group_profile = models.GroupProfile({
                "name": NEW_GROUP_NAME
            })
            new_group = models.Group({
                "profile": new_group_profile
            })

            # Create 2nd group
            temp_group, _, err = await test_client.create_group(new_group)
            assert err is None

            # Add new group target to role and remove original
            _, err = await test_client.add_group_target_to_role(
                user.id, user_role.id, temp_group.id)
            assert err is None
            _, err = await test_client.remove_group_target_from_role(
                user.id, user_role.id, group.id)
            assert err is None

        finally:
            errors = []
            # Deactivate, then delete created user
            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await test_client.deactivate_or_delete_user(user.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            # Delete groups created
            try:
                _, err = await test_client.delete_group(group.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await test_client.delete_group(temp_group.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)

            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_user_pagination(self, fs):
        # Instantiate Mock Client
        test_client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create 2 user profiles
        first_user_profile = models.UserProfile()
        first_user_profile.first_name = "John"
        first_user_profile.last_name = "Doe-Paginate-1"
        first_user_profile.email = "John.Doe-Paginate-1@example.com"
        first_user_profile.login = "John.Doe-Paginate-1@example.com"

        second_user_profile = models.UserProfile()
        second_user_profile.first_name = "John"
        second_user_profile.last_name = "Doe-Paginate-2"
        second_user_profile.email = "John.Doe-Paginate-2@example.com"
        second_user_profile.login = "John.Doe-Paginate-2@example.com"

        first_create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": first_user_profile
        })
        second_create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": second_user_profile
        })

        try:
            # Create Query Parameters and Create 2 Users
            query_params_create = {"activate": "False"}
            user1, _, err = await test_client.create_user(
                first_create_user_req, query_params_create)
            assert err is None
            user2, _, err = await test_client.create_user(
                second_create_user_req, query_params_create)
            assert err is None

            # Create Query Parameters and List Users
            query_params_limit = {"limit": "1"}  # Retrieve one user
            usr_list, resp, err = await test_client.list_users(query_params_limit)
            assert err is None
            assert len(usr_list) == 1
            assert isinstance(usr_list[0], models.User)
            assert resp.has_next()  # ensure next can be determined

            # Retrieve next page of results
            second_usr_list, err = await resp.next()
            assert err is None
            assert len(second_usr_list) == 1
            assert isinstance(second_usr_list[0], models.User)

            # Deactivate & delete both users
            _, err = await test_client.deactivate_or_delete_user(user1.id)
            assert err is None
            _, err = await test_client.deactivate_or_delete_user(user1.id)
            assert err is None

            _, err = await test_client.deactivate_or_delete_user(user2.id)
            assert err is None
            _, err = await test_client.deactivate_or_delete_user(user2.id)
            assert err is None

            # Ensure deletion
            found_user, _, err = await test_client.get_user(user1.id)
            assert found_user is None
            assert isinstance(err, OktaAPIError)
        finally:
            try:
                _, err = await test_client.deactivate_or_delete_user(user1.id)
            except Exception:
                pass
            try:
                _, err = await test_client.deactivate_or_delete_user(user1.id)
            except Exception:
                pass
            try:
                _, err = await test_client.deactivate_or_delete_user(user2.id)
            except Exception:
                pass
            try:
                _, err = await test_client.deactivate_or_delete_user(user2.id)
            except Exception:
                pass

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_user_subscriptions(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        users, _, err = await client.list_users()
        assert err is None
        user = users[0]

        resp, _, err = await client.list_user_subscriptions(user.id)
        assert len(resp) > 0
        for item in resp:
            assert isinstance(item, models.Subscription)
            assert isinstance(item.status, models.SubscriptionStatus)
            assert isinstance(item.notification_type, models.NotificationType)
